/* machine/stdlib.h

  Copyright 2005 Red Hat, Inc.

  This file is part of Cygwin.

  This software is a copyrighted work licensed under the terms of the
  Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
  details. */

#ifndef _MACHSTDLIB_H_
#define _MACHSTDLIB_H_
#ifdef __cplusplus
extern "C" {
#endif

char *mkdtemp (char *);

#ifdef __cplusplus
}
#endif
#endif  /* _MACHSTDLIB_H_ */
