/* sys/ipc.h

   Copyright 2001, 2002, 2003 Red Hat, Inc.

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

/* sys/ipc.h header file for Cygwin.  */

#ifndef _SYS_IPC_H
#define _SYS_IPC_H

#include <cygwin/ipc.h>

#endif /* _SYS_IPC_H */
