/* sys/xattr.h

   Copyright 2011 Red Hat, Inc.

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

/* sys/xattr.h header file for Cygwin.  */

#ifndef _SYS_XATTR_H
#define _SYS_XATTR_H

#include <attr/xattr.h>

#endif /* _SYS_XATTR_H */
