#include <winver.h>
#include <cygwin/version.h>

#define STRINGIFY1(x) #x
#define STRINGIFY(x) STRINGIFY1(x)

#define CYGWIN_DLL_NAME CYGWIN_VERSION_DLL_IDENTIFIER STRINGIFY(.dll)

#define CYGWIN_API_VERSION STRINGIFY(CYGWIN_VERSION_API_MAJOR) "." \
			  STRINGIFY(CYGWIN_VERSION_API_MINOR)

#define CYGWIN_BUILD_DATE_TIME	STRINGIFY(CYGWIN_BUILD_DATE) " " \
				STRINGIFY(CYGWIN_BUILD_TIME)

VS_VERSION_INFO VERSIONINFO
  FILEVERSION    CYGWIN_VERSION_DLL_MAJOR,CYGWIN_VERSION_DLL_MINOR,0,0
  PRODUCTVERSION CYGWIN_VERSION_DLL_MAJOR,CYGWIN_VERSION_DLL_MINOR,0,0
  FILEFLAGSMASK  0x3fL
#ifdef DEBUGGING
  FILEFLAGS      VS_FF_DEBUG
#else
  FILEFLAGS      0x0L
#endif
  FILEOS         VOS__WINDOWS32
  FILETYPE       VFT_DLL
  FILESUBTYPE    0x0L
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904B0"
    BEGIN
      VALUE "CompanyName", "Red Hat"
      VALUE "FileDescription", "Cygwin POSIX Emulation DLL"
      VALUE "FileVersion", STRINGIFY(CYGWIN_VERSION)
      VALUE "InternalName", CYGWIN_DLL_NAME
      VALUE "LegalCopyright", "Copyright \251 Red Hat, Inc. 1996-" STRINGIFY(CYGWIN_BUILD_YEAR)
      VALUE "OriginalFilename", CYGWIN_DLL_NAME
      VALUE "ProductName", "Cygwin"
      VALUE "ProductVersion", STRINGIFY(CYGWIN_VERSION)
      VALUE "APIVersion", CYGWIN_API_VERSION
      VALUE "SharedMemoryVersion", STRINGIFY(CYGWIN_VERSION_SHARED_DATA)
      VALUE "RegistryKey", CYGWIN_INFO_CYGWIN_REGISTRY_NAME
      VALUE "BuildDate", CYGWIN_BUILD_DATE_TIME
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
      VALUE "Translation", 0x409, 1200
  END
END
