/* sys/mtio.h

   Copyright 1999, 2000, 2001 Red Hat, Inc.

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

/* sys/mtio.h header file for Cygwin.  */

#ifndef _SYS_MTIO_H
#define _SYS_MTIO_H

#include <cygwin/mtio.h>

#endif /* _SYS_MTIO_H */
