/* sys/kd.h

   Copyright 2006 Red Hat, Inc.

   Written by Kazuhiro Fujieda <fujieda@jaist.ac.jp>

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

/* sys/kd.h header file for Cygwin.  */

#ifndef _SYS_KD_H
#define _SYS_KD_H

#include <cygwin/kd.h>

#endif /* _SYS_KD_H */
