/* paths.h

   Copyright 2000, 2001, 2002, 2003, 2010, 2011 Red Hat, Inc.

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

#ifndef	_PATHS_H_
#define _PATHS_H_

#define _PATH_BSHELL    "/bin/sh"
#define _PATH_CSHELL	"/bin/csh"
#define _PATH_DEFPATH	"/bin"
#define _PATH_DEV	"/dev/"
#define _PATH_DEVNULL	"/dev/null"
#define _PATH_LASTLOG	"/var/log/lastlog"
#define _PATH_MAILDIR	"/var/spool/mail/"
#define _PATH_MAN	"/usr/share/man"
#define _PATH_MEM	"/dev/mem"
#define _PATH_MNTTAB	"/etc/fstab"
#define _PATH_MOUNTED	"/etc/mtab"
#define _PATH_SHELLS	"/etc/shells"
#define _PATH_STDPATH	"/bin:/usr/sbin:/sbin"
#define _PATH_TMP	"/tmp/"
#define _PATH_TTY	"/dev/tty"
#define _PATH_UTMP	"/var/run/utmp"
#define _PATH_VARRUN	"/var/run/"
#define _PATH_VARTMP	"/var/tmp/"
#define _PATH_VI        "/bin/vi"
#define _PATH_WTMP	"/var/log/wtmp"

#endif /* _PATHS_H_ */
