/* wait.h

   Copyright 2008 Red Hat, Inc.

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

#ifndef _WAIT_H
#define _WAIT_H

#include <sys/wait.h>

#endif /* _WAIT_H */
