/* cygwin/kd.h

   Copyright 2006 Red Hat Inc.
   Written by Kazuhiro Fujieda <fujieda@jaist.ac.jp>

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

#ifndef _CYGWIN_KD_H_
#define _CYGWIN_KD_H_

#define KDGKBMETA 0x4b62
#define KDSKBMETA 0x4b63
#define K_METABIT 0x03
#define K_ESCPREFIX 0x04

#endif /* _CYGWIN_KD_H_ */
