/* premain2.c

   Copyright 2000, 2013 Red Hat, Inc.

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

void
cygwin_premain2(int argc __attribute__ ((unused)),
		char **argv __attribute__ ((unused)))
{
}
