/* Define if DEBUGGING support is requested.  */
#undef DEBUGGING

/* Define if GCC supports builtin memset.  */
#undef HAVE_BUILTIN_MEMSET

/* Define if MALLOC_DEBUGGING support is requested.  */
#undef MALLOC_DEBUG

/* Define if using new vfork functionality. */
#undef NEWVFORK
